# fiscaldataapi
This repository contains a Python wrapper to easily retrieve data from the Fiscal Data (US Treasury) official API in pandas format.

## Overview
The U.S. Department of the Treasury is building a suite of open-source tools to deliver standardized information about federal finances to the public. We are working to centralize publicly available financial data, and this website will include datasets from the Fiscal Service on topics including debt, revenue, spending, interest rates, and savings bonds.
## Requirements
- Python 3.9 or higher.
- Requests
- Pandas

## Installation
```terminal
pip install fiscaldataapi
```

## Get Fiscal Data
### get_fiscaldata
Get FiscalData table from Treasury series data.

```python
from fiscaldataapi import get_fiscaldata
df = get_fiscaldata(database, **kwargs)
```

#### Parameters
| Parameter  | Type  | Description                |
|------------|-------|----------------------------|
| `database` | `str` | Table name of the endpoint |

#### Keyword Arguments
| Parameter     | Type         | Description                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                              |
|---------------|--------------|----------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------|
| `fields`      | `str`        | The fields parameter allows you to select which field(s) should be included. If desired fields are not specified, all fields will be returned. When a file name passed to the fields parameter is not available for the endpoint accessed, an error will occur.                                                                                                                                                                                                                                                                                          |
| `filter`      | `str`        | Filters are used to view a subset of the data based on specific criteria. For example, you may want to find data that falls within a certain date range, or only show records which contain a value larger than a certain threshold.  Use a colon at the end of a filter parameter to pass a value or list of values. For lists passed as filter criteria, use a comma-separated list within parentheses. Filter for specific dates using the format YYYY-MM-DD. When no filters are provided, the default response will return all fields and all data. |
| `sort`        | `str`        | The sort parameter allows a user to sort a field in ascending (least to greatest) or descending (greatest to least) order. The sort parameter accepts a comma-separated list of field names. When no sort parameter is specified, the default is to sort by the first column listed.                                                                                                                                                                                                                                                                     |
| `page_size`   | `str or int` | The page size will set the number of rows that are returned on a request. When no page size parameter is specified, the default response is 100.                                                                                                                                                                                                                                                                                                                                                                                                         |
| `page_number` | `str or int` | The page number will set the index for the pagination, starting at 1. When no page number is specified, the default response is 1.                                                                                                                                                                                                                                                                                                                                                                                                                       |

#### Returns
- pd.DataFrame: Dataframe containing the FiscalData endpoint table data.


### info_fiscaldata
Get information of a FiscalData table.

```python
from fiscaldataapi import info_fiscaldata
df = info_fiscaldata(database)
```

#### Parameters
| Parameter  | Type  | Description                |
|------------|-------|----------------------------|
| `database` | `str` | Table name of the endpoint |

#### Returns
- tuple(dict, int, int): Dictionary with the format of the table columns, the number of pages of table data (each page has up to 100 rows) and number of total rows of table data.



## Table of tables and endpoints
| Dataset                                                              | Table Name                                                                                           | Endpoint                                                                                     | Endpoint Description                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                            |
|----------------------------------------------------------------------|------------------------------------------------------------------------------------------------------|----------------------------------------------------------------------------------------------|-----------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------|
| 120 Day Delinquent Debt Referral Compliance Report                   | 120 Day Delinquent Debt Referral Compliance Report                                                   | `/v2/debt/tror/data_act_compliance`                                                          | The 120 Day Delinquent Debt Referral Compliance Report provides access to tracking and benchmarking compliance with the requirements of a key provision of the Digital Accountability and Transparency Act of 2014 (the DATA Act). This table provides quick insights into federal agency compliance rates, as well as information on the number of eligible debts and debts referred or not referred each quarter, beginning in Fiscal Year 2016.                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                              |
| Accrual Savings Bonds Redemption Tables (Discontinued)               | Redemption Tables                                                                                    | `/v2/accounting/od/redemption_tables`                                                        | The Redemption Tables dataset contains monthly tables that list the redemption value, interest earned, and yield of accrual savings bonds purchased since 1941. Each monthly report lists the redemption value of all bonds at the time of publication. Investors and bond owners can use this dataset as an easy and understandable reference to know the redemption value of the bonds they hold.                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                             |
| Advances to State Unemployment Funds (Social Security Act Title XII) | Advances to State Unemployment Funds (Social Security Act Title XII)                                 | `/v2/accounting/od/title_xii`                                                                | Monthly balances for securities outstanding and principal outstanding for State and Local Government Series (SLGS) securities.                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                  |
| Average Interest Rates on U.S. Treasury Securities                   | Average Interest Rates on U.S. Treasury Securities                                                   | `/v2/accounting/od/avg_interest_rates`                                                       | Average interest rates for marketable and non-marketable securities.                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                            |
| Daily Treasury Statement (DTS)                                       | Operating Cash Balance                                                                               | `/v1/accounting/dts/operating_cash_balance`                                                  | This table represents the Treasury General Account balance. Additional detail on changes to the Treasury General Account can be found in the Deposits and Withdrawals of Operating Cash table. All figures are rounded to the nearest million.                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                  |
| Daily Treasury Statement (DTS)                                       | Deposits and Withdrawals of Operating Cash                                                           | `/v1/accounting/dts/deposits_withdrawals_operating_cash`                                     | This table represents deposits and withdrawals from the Treasury General Account. A summary of changes to the Treasury General Account can be found in the Operating Cash Balance table. All figures are rounded to the nearest million.                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                        |
| Daily Treasury Statement (DTS)                                       | Public Debt Transactions                                                                             | `/v1/accounting/dts/public_debt_transactions`                                                | This table represents the issues and redemption of marketable and nonmarketable securities. All figures are rounded to the nearest million.                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                     |
| Daily Treasury Statement (DTS)                                       | Adjustment of Public Debt Transactions to Cash Basis                                                 | `/v1/accounting/dts/adjustment_public_debt_transactions_cash_basis`                          | This table represents cash basis adjustments to the issues and redemptions of Treasury securities in the Public Debt Transactions table. All figures are rounded to the nearest million.                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                        |
| Daily Treasury Statement (DTS)                                       | Debt Subject to Limit                                                                                | `/v1/accounting/dts/debt_subject_to_limit`                                                   | This table represents the breakdown of total public debt outstanding as it relates to the statutory debt limit. All figures are rounded to the nearest million.                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                 |
| Daily Treasury Statement (DTS)                                       | Inter-Agency Tax Transfers                                                                           | `/v1/accounting/dts/inter_agency_tax_transfers`                                              | This table represents the breakdown of inter-agency tax transfers within the federal government. All figures are rounded to the nearest million.                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                |
| Daily Treasury Statement (DTS)                                       | Income Tax Refunds Issued                                                                            | `/v1/accounting/dts/income_tax_refunds_issued`                                               | This table represents the breakdown of tax refunds by recipient (individual vs business) and type (check vs electronic funds transfer). Tax refunds are also represented as withdrawals in the Deposits and Withdrawals of Operating Cash table. All figures are rounded to the nearest million. As of February 14, 2023, Table VI Income Tax Refunds Issued was renamed to Table V Income Tax Refunds Issued within the published report.                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                      |
| Daily Treasury Statement (DTS)                                       | Federal Tax Deposits                                                                                 | `/v1/accounting/dts/federal_tax_deposits`                                                    | This table represents the breakdown of taxes that are received by the federal government. Federal taxes received are represented as deposits in the Deposits and Withdrawals of Operating Cash table. All figures are rounded to the nearest million.                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                           |
| Daily Treasury Statement (DTS)                                       | Short-Term Cash Investments                                                                          | `/v1/accounting/dts/short_term_cash_investments`                                             | This table represents the amount Treasury has in short-term cash investments. Deposits and withdrawals of short-term cash investments are also represented in the Deposits and Withdrawals of Operating Cash table. This program was suspended indefinitely in 2008. All figures are rounded to the nearest million. As of February 14, 2023, Table V Short Term Cash Investments will no longer be updated and removed from the published report. The historical data will remain available.                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   |
| Debt to the Penny                                                    | Debt to the Penny                                                                                    | `/v2/accounting/od/debt_to_penny`                                                            | Outstanding U.S. debt on a daily basis.                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                         |
| Electronic Securities Transactions                                   | Sales                                                                                                | `/v1/accounting/od/securities_sales`                                                         | Securities Issued in TreasuryDirect: Sales provides gross sales, returns, net sales, and count of securities for marketable securities and savings bonds issued in TreasuryDirect. Securities are available by security class: bills, notes, bonds, and TIPS for marketable securities and Series E, EE, and I for savings bonds.                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                               |
| Electronic Securities Transactions                                   | Sales by Term                                                                                        | `/v1/accounting/od/securities_sales_term`                                                    | Securities Issued in TreasuryDirect: Sales by Term provides the count and dollar value of marketable securities and savings bonds issued in TreasuryDirect by term. Securities are available by security class: bills, notes, bonds, and Treasury Inflation-Protected Securities (TIPS) for marketable securities and Series EE and I for savings bonds.                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                        |
| Electronic Securities Transactions                                   | Transfers of Marketable Securities                                                                   | `/v1/accounting/od/securities_transfers`                                                     | Securities Issued In TreasuryDirect: Transfers of Marketable Securities provides the count and dollar value for marketable securities transferred to TreasuryDirect from legacy Treasury Direct, legacy holding system, and commercial book entry.                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                              |
| Electronic Securities Transactions                                   | Conversions of Paper Savings Bonds                                                                   | `/v1/accounting/od/securities_conversions`                                                   | Securities Issued in TreasuryDirect: Conversions of Paper Savings Bonds provides the count and dollar value of paper savings bonds converted to electronic savings bonds in TreasuryDirect.                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                     |
| Electronic Securities Transactions                                   | Redemptions                                                                                          | `/v1/accounting/od/securities_redemptions`                                                   | Securities Issued in TreasuryDirect: Redemptions provides the count and dollar value for marketable securities and savings bonds redeemed in TreasuryDirect. Securities are available by security class: bills, notes, bonds, and Treasury Inflation-Protected Securities (TIPS) for marketable securities and Series E, EE, and I for savings bonds.                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                           |
| Electronic Securities Transactions                                   | Outstanding                                                                                          | `/v1/accounting/od/securities_outstanding`                                                   | Securities Issued in TreasuryDirect: Outstanding provides the dollar value for marketable securities and savings bonds outstanding in TreasuryDirect. Securities are available by security class: bills, notes, bonds, and TIPS for marketable securities and Series E, EE, and I for savings bonds. Does not include outstanding dollar value of certificates of indebtedness. That data can be retrieved from the Securities Issued in TreasuryDirect: Certificates of Indebtedness table.                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                    |
| Electronic Securities Transactions                                   | Certificates of Indebtedness                                                                         | `/v1/accounting/od/securities_c_of_i`                                                        | Securities Issued In TreasuryDirect: Certificates of Indebtedness provides the dollar amount of outstanding certificates of indebtedness issued in TreasuryDirect.                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                              |
| Electronic Securities Transactions                                   | Accounts                                                                                             | `/v1/accounting/od/securities_accounts`                                                      | Securities Issued in TreasuryDirect: Accounts provides the count of unrestricted primary accounts and funded accounts in TreasuryDirect.                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                        |
| Federal Borrowings Program: Interest on Uninvested Funds             | Federal Borrowings Program: Interest on Uninvested Funds                                             | `/v2/accounting/od/interest_uninvested`                                                      | A quarterly report containing interest balances associated with Treasury's Credit Reform: Interest Paid on Uninvested Funds account.                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                            |
| Federal Borrowings Program: Summary General Ledger Balances Report   | Summary General Ledger Borrowing Balances                                                            | `/v1/accounting/od/fbp_gl_borrowing_balances`                                                | The Summary General Ledger Borrowing Balances table provides Treasury Loans Receivable, Capitalized Interest Receivable, Interest Receivable, Interest Revenue, Gain, and Loss balances associated with every corresponding borrowing agency expenditure Treasury Account Symbol.                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                               |
| Federal Borrowings Program: Summary General Ledger Balances Report   | Summary General Ledger Repayable Advance Balances                                                    | `/v1/accounting/od/fbp_gl_repay_advance_balances`                                            | The Summary General Ledger Repayable Advance Balances table provides Treasury Loans Receivable, Capitalized Interest Receivable, Interest Receivable, Interest Revenue, Gain, and Loss balances associated with every corresponding borrowing agency expenditure Treasury Account Symbol. The *0002 represents the extended unemployment compensation account within the unemployment trust fund and the *0003 represents the federal unemployment account within the unemployment trust fund.                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                  |
| Federal Credit Similar Maturity Rates                                | Federal Credit Similar Maturity Rates                                                                | `/v1/accounting/od/federal_maturity_rates`                                                   | The Federal Credit Similar Maturity Rates table provides the annual rates for outstanding fixed-rate Treasury securities, categorized on the basis of their remaining maturity.                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                 |
| Federal Investments Program: Interest Cost by Fund                   | Federal Investments Program: Interest Cost by Fund                                                   | `/v2/accounting/od/interest_cost_fund`                                                       | A monthly summary report containing premiums, discounts, accrued interest collected, interest payments, and inflation compensation for all accounts invested in Government Account Series (GAS) Securities, as well as premium and discount amortization for accounts invested in zero-coupon bonds (ZCBs).                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                     |
| Federal Investments Program: Principal Outstanding                   | Principal Outstanding                                                                                | `/v1/accounting/od/fip_principal_outstanding_table1`                                         | This table provides information on shares/par amounts by security sectors for a specific account statement number.                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                              |
| Federal Investments Program: Principal Outstanding                   | Total Outstanding Inflation Compensation                                                             | `/v1/accounting/od/fip_principal_outstanding_table2`                                         | This table provides inflation compensation (accrued) totals for specific accounts.                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                              |
| Federal Investments Program: Statement of Account                    | CARS Reporting                                                                                       | `/v1/accounting/od/fip_statement_of_account_table1`                                          | This table provides information on investments and earnings increases and decreases for a specific account statement number.                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                    |
| Federal Investments Program: Statement of Account                    | Account Position Summary                                                                             | `/v1/accounting/od/fip_statement_of_account_table2`                                          | This table provides beginning and ending balances for specific categories for a specific account statement number.                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                              |
| Federal Investments Program: Statement of Account                    | Transaction Detail                                                                                   | `/v1/accounting/od/fip_statement_of_account_table3`                                          | This table provides transition details for a specific account statement number.                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                 |
| Financial Report of the U.S. Government                              | Statements of Net Cost                                                                               | `/v2/accounting/od/statement_net_cost`                                                       | The Statement of Net Cost presents the net cost of government operations, including the operations related to funds from dedicated collections. Costs and earned revenues are categorized on the Statement of Net Cost by significant entity, providing greater accountability by showing the relationship of entities' net costs to the government-wide net cost. Net cost is computed by subtracting earned revenue from gross cost, adjusted by the total gain or loss from changes in assumptions.                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                          |
| Financial Report of the U.S. Government                              | Statements of Operations and Changes in Net Position                                                 | `/v1/accounting/od/net_position`                                                             | The Statement of Operations and Changes in Net Position (SOCNP) reports the results of government operations, net operating costs, which includes the results of operations for funds from dedicated collections. The statement includes non-exchange revenues, which are inflows of resources to the government that the government demands or that it receives by donations are identified, such as taxes and duties. The net operating cost equals revenue less net cost of government operations (that flows from the Statement of Net Cost) adjusted by intra-governmental transfers and unmatched transactions and balances. The change in net position between the beginning and end of the fiscal year reflects adjustments to beginning net position and net operating costs.                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                          |
| Financial Report of the U.S. Government                              | Reconciliations of Net Operating Cost and Budget Deficit                                             | `/v1/accounting/od/reconciliations`                                                          | The Reconciliation of Net Operating Cost and Budget Deficit statement covers the reconciliation of the results of operations (net operating cost from the Statement of Operations and Changes in Net Position) to the budget deficit. Receipts and outlays in the budget are measured primarily on a cash basis as opposed to the accrual basis of accounting used in the Financial Report. The premise of the reconciliation is that transaction data is shared between the accrual and budgetary cash accounting basis. These statements begin with net operating cost and report activities where the basis of accounting for the components of net operating cost and the budget deficit differ, as well as adjustments to beginning net position and other reconciling items.                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                              |
| Financial Report of the U.S. Government                              | Statements of Changes in Cash Balance from Budget and Other Activities                               | `/v1/accounting/od/cash_balance`                                                             | The Statement of Changes in Cash Balance from Budget and Other Activities covers cash flow from budget activities, adjustments from non-cash outlays included in the budget, cash flow from activities not included in the budget. It details how the annual budget deficit relates to the change in the government's cash, other monetary assets, federal debt, and interest payable. It also explains how the budget deficits were financed.                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                  |
| Financial Report of the U.S. Government                              | Balance Sheets                                                                                       | `/v2/accounting/od/balance_sheets`                                                           | The Balance Sheet provides an overview of the government's assets, liabilities, and net position to present a more comprehensive report of the government's financial position. The most significant assets that are reported on the balance sheets are loans receivable; property, plant, and equipment; inventories and related property; and cash or other monetary assets. The most significant liabilities reported on the balance sheets are federal debt securities held by the public and accrued interest, and federal employee and veteran benefits payable. The net position is the residual difference between assets and liabilities and is the cumulative results of operations since inception.                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                  |
| Financial Report of the U.S. Government                              | Statements of Long-Term Fiscal Projections                                                           | `/v1/accounting/od/long_term_projections`                                                    | The Statement of Long-Term Fiscal Projections displays the present value of 75-year projections by major category of receipts and non-interest spending. It aids in the assessment of the federal government's financial condition and how it has changed during the year and may change in the future. It also helps in evaluating whether future budgetary resources will be sufficient to sustain public services and to meet obligations when due, assuming that current policy for federal government public services and taxation is continued without change. This statement presents data as a percent of GDP to help readers assess whether current fiscal policy is sustainable.                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                      |
| Financial Report of the U.S. Government                              | Statements of Social Insurance                                                                       | `/v1/accounting/od/social_insurance`                                                         | The Statement of Social Insurance (SOSI) provides long-range estimates that can be used to assess the financial condition of the most significant U.S. social insurance programs: Social Security, Medicare, Railroad Retirement, and Black Lung. These programs are administered by the Social Security Administration, the Department of Health & Human Services, the Railroad Retirement Board, and the Department of Labor, respectively. SOSI displays 75-year projections (except for Black Lung) by major social insurance program, illustrating the relationship between the estimated receipts and expenditures of those programs.                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                     |
| Financial Report of the U.S. Government                              | Statements of Changes in Social Insurance Amounts                                                    | `/v1/accounting/od/insurance_amounts`                                                        | The Statement of Changes in Social Insurance Amounts (SCSIA) shows reasons for changes in the Statement of Social Insurance (SOSI) estimates between one current valuation period and the prior valuation period. It details the reasons for the change by social insurance program, including changes in factors such as (1) demographic data, assumptions, and methods; (2) policy; and (3) economic and other health care assumptions.                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                       |
| FRN Daily Indexes                                                    | FRN Daily Indexes                                                                                    | `/v1/accounting/od/frn_daily_indexes`                                                        | The FRN Daily Indexes dataset provides data on Floating Rate Notes. For floating rate notes, the index is the highest accepted discount rate on 13-week bills determined by Treasury auctions of those securities. We auction the 13-week Treasury bill every week, so the index rate of an FRN is reset every week. The FRN Daily Indexes provide information for specific CUSIPs, accrual periods, daily indexes, daily interest accrual rates, spread, and interest payment periods.                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                         |
| Gift Contributions to Reduce the Public Debt                         | Gift Contributions to Reduce the Public Debt                                                         | `/v2/accounting/od/gift_contributions`                                                       | Summary of gifts donated to the United States Government to reduce debt held by the public.                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                     |
| Historical Debt Outstanding                                          | Historical Debt Outstanding                                                                          | `/v2/accounting/od/debt_outstanding`                                                         | U.S. debt outstanding at the end of each fiscal year.                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                           |
| Historical Qualified Tax Credit Bond Interest Rates                  | Historical Qualified Tax Credit Bond Interest Rates                                                  | `/v2/accounting/od/qualified_tax`                                                            | Historical interest rate, term to maturity, and permitted sinking fund yield for qualified tax credit bonds reported daily through January 30, 2018.                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                            |
| Interest Expense on the Public Debt Outstanding                      | Interest Expense on the Public Debt Outstanding                                                      | `/v2/accounting/od/interest_expense`                                                         | The Interest Expense on the Debt Outstanding is a monthly summary of the cost of interest on U.S. debt, including U.S. Treasury notes and bonds, foreign and domestic series certificates of indebtedness, notes and bonds, savings bonds, State and Local Government series (SLGS) and other special purpose securities.                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                       |
| Judgment Fund: Annual Report to Congress                             | Judgment Fund: Annual Report to Congress                                                             | `/v2/payments/jfics/jfics_congress_report`                                                   | This table provides a listing of payments made through the Judgment Fund including the amount paid out, judgment type, legal representatives, agencies involved, and associated costs.                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                          |
| Monthly State and Local Government Series (SLGS) Securities Program  | Monthly State and Local Government Series (SLGS) Securities Program                                  | `/v2/accounting/od/slgs_statistics`                                                          | This table shows what states and territories are borrowing from the federal Unemployment Trust Fund in order to pay out unemployment benefits.                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                  |
| Monthly Treasury Statement (MTS)                                     | Summary of Receipts, Outlays, and the Deficit/Surplus of the U.S. Government                         | `/v1/accounting/mts/mts_table_1`                                                             | This summary table shows the total amount of receipts and outlays and the amount of the budget surplus/deficit by month for the current and prior fiscal years. This table includes total and subtotal rows that should be excluded when aggregating data. Some rows represent elements of the dataset's hierarchy, but are not assigned values. The classification_id for each of these elements can be used as the parent_id for underlying data elements to calculate their implied values. Subtotal rows are available to access this same information.                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                     |
| Monthly Treasury Statement (MTS)                                     | Summary of Budget and Off-Budget Results and Financing of the U.S. Government                        | `/v1/accounting/mts/mts_table_2`                                                             | This summary table shows the on-budget and off-budget receipts and outlays, the on-budget and off-budget surplus/deficit, and the means of financing the budget surplus/deficit. The table also shows the budgeted amounts estimated in the President's Budget for the current fiscal year and next fiscal year for each item on the table. This table includes total and subtotal rows that should be excluded when aggregating data. Some rows represent elements of the dataset's hierarchy, but are not assigned values. The classification_id for each of these elements can be used as the parent_id for underlying data elements to calculate their implied values. Subtotal rows are available to access this same information.                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                         |
| Monthly Treasury Statement (MTS)                                     | Summary of Receipts and Outlays of the U.S. Government                                               | `/v1/accounting/mts/mts_table_3`                                                             | This summary table shows, for Budget Receipts, the total amount of activity for the current month, the current fiscal year-to-date, the comparable prior period year-to-date and the budgeted amount estimated for the current fiscal year for various types of receipts (i.e. individual income tax, corporate income tax, etc.). The Budget Outlays section of the table shows the total amount of activity for the current month, the current fiscal year-to-date, the comparable prior period year-to-date and the budgeted amount estimated for the current fiscal year for agencies of the federal government. The table also shows the amounts for the budget/surplus deficit categorized as listed above. This table includes total and subtotal rows that should be excluded when aggregating data. Some rows represent elements of the dataset's hierarchy, but are not assigned values. The classification_id for each of these elements can be used as the parent_id for underlying data elements to calculate their implied values. Subtotal rows are available to access this same information.                                                                                                                                                                                                                                                   |
| Monthly Treasury Statement (MTS)                                     | Receipts of the U.S. Government                                                                      | `/v1/accounting/mts/mts_table_4`                                                             | This table shows the gross receipts, refunds and net receipts for the current month, the current fiscal year-to-date and the prior fiscal year-to-date for the various receipts of the federal government. This table includes total and subtotal rows that should be excluded when aggregating data. Some rows represent elements of the dataset's hierarchy, but are not assigned values. The classification_id for each of these elements can be used as the parent_id for underlying data elements to calculate their implied values. Subtotal rows are available to access this same information.                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                          |
| Monthly Treasury Statement (MTS)                                     | Outlays of the U.S. Government                                                                       | `/v1/accounting/mts/mts_table_5`                                                             | This table shows the gross outlays, applicable receipts and net outlays for the current month, current fiscal year-to-date and prior fiscal year-to-date by various agency programs accounted for in the budget of the federal government. This table includes total and subtotal rows that should be excluded when aggregating data. Some rows represent elements of the dataset's hierarchy, but are not assigned values. The classification_id for each of these elements can be used as the parent_id for underlying data elements to calculate their implied values. Subtotal rows are available to access this same information.                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                          |
| Monthly Treasury Statement (MTS)                                     | Means of Financing the Deficit or Disposition of Surplus by the U.S. Government                      | `/v1/accounting/mts/mts_table_6`                                                             | This table shows the net transactions for the current month, and the current and prior fiscal year-to-date, as well as account balances for the beginning of the current fiscal year and current accounting month and the close of the current accounting month. This activity is related to the means used to finance the budget deficit or to dispose of a budget surplus. An asset account would represent an asset to the United States Government, for example United States Treasury Operating Cash. A liability account would represent a liability to the United States Government, for example Borrowing from the Public. This table includes total and subtotal rows that should be excluded when aggregating data. Some rows represent elements of the dataset's hierarchy, but are not assigned values. The classification_id for each of these elements can be used as the parent_id for underlying data elements to calculate their implied values. Subtotal rows are available to access this same information.                                                                                                                                                                                                                                                                                                                                  |
| Monthly Treasury Statement (MTS)                                     | Analysis of Change in Excess of Liabilities of the U.S. Government                                   | `/v1/accounting/mts/mts_table_6a`                                                            | This table is a subsidiary table for Means of Financing the Deficit or Disposition of Surplus by the U.S. Government providing a detailed view of the Change in Excess of Liabilities. This table includes total and subtotal rows that should be excluded when aggregating data. Some rows represent elements of the dataset's hierarchy, but are not assigned values. The classification_id for each of these elements can be used as the parent_id for underlying data elements to calculate their implied values. Subtotal rows are available to access this same information.                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                              |
| Monthly Treasury Statement (MTS)                                     | Securities Issued by Federal Agencies Under Special Financing Authorities                            | `/v1/accounting/mts/mts_table_6b`                                                            | This table is a subsidiary table for Means of Financing the Deficit or Disposition of Surplus by the U.S. Government providing a detailed view of the transactions labelled, Agency Securities, Issued Under Special Financing Authorities. Special financing authorities include financing that is established by legislation under special or unique circumstances and for a specific purpose. This table includes total and subtotal rows that should be excluded when aggregating data. Some rows represent elements of the dataset's hierarchy, but are not assigned values. The classification_id for each of these elements can be used as the parent_id for underlying data elements to calculate their implied values. Subtotal rows are available to access this same information.                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                    |
| Monthly Treasury Statement (MTS)                                     | Federal Agency Borrowing Financed Through the Issue of Treasury Securities                           | `/v1/accounting/mts/mts_table_6c`                                                            | This table is a subsidiary table for Means of Financing the Deficit or Disposition of Surplus by the U.S. Government, providing a detailed view of transactions and account balances for agency programs that borrow from the United States Treasury or from the Federal Financing Bank. This table includes total and subtotal rows that should be excluded when aggregating data. Some rows represent elements of the dataset's hierarchy, but are not assigned values. The classification_id for each of these elements can be used as the parent_id for underlying data elements to calculate their implied values. Subtotal rows are available to access this same information.                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                            |
| Monthly Treasury Statement (MTS)                                     | Investments of Federal Government Accounts in Federal Securities                                     | `/v1/accounting/mts/mts_table_6d`                                                            | This table is a subsidiary table for Means of Financing the Deficit or Disposition of Surplus by the U.S. Government providing a detailed view of federal funds and trust funds that are invested in Government Account Series (GAS) securities. Federal funds include general funds, special funds, and revolving funds (public enterprise revolving funds, intragovernmental revolving funds, and credit financing accounts). A trust fund is a type of account, designated by law, for receipts or offsetting receipts dedicated to specific purposes and the expenditure of these receipts. This table includes total and subtotal rows that should be excluded when aggregating data. Some rows represent elements of the dataset's hierarchy, but are not assigned values. The classification_id for each of these elements can be used as the parent_id for underlying data elements to calculate their implied values. Subtotal rows are available to access this same information.                                                                                                                                                                                                                                                                                                                                                                     |
| Monthly Treasury Statement (MTS)                                     | Guaranteed and Direct Loan Financing, Net Activity                                                   | `/v1/accounting/mts/mts_table_6e`                                                            | This table is a subsidiary table for Means of Financing the Deficit or Disposition of Surplus by the U.S. Government providing a detailed view of all direct and guaranteed loan financing for federal credit programs under the Credit Reform Act of 1990. Guaranteed loan financing is issuing any debt obligation with a guarantee, insurance, or other pledge that payment of all or a part of the principal or interest will be made to the lender. This table applies to lending to non-federal borrowers by non-federal lenders that carries some form of guarantee by the federal government. Exceptions include the insurance of deposits, shares, or other withdrawable accounts in financial institutions. This table includes total and subtotal rows that should be excluded when aggregating data. Some rows represent elements of the dataset's hierarchy, but are not assigned values. The classification_id for each of these elements can be used as the parent_id for underlying data elements to calculate their implied values. Subtotal rows are available to access this same information.                                                                                                                                                                                                                                               |
| Monthly Treasury Statement (MTS)                                     | Receipts and Outlays of the U.S. Government by Month                                                 | `/v1/accounting/mts/mts_table_7`                                                             | This table shows the receipts and outlays of the United States Government by month for the current fiscal year, up to and including the current accounting month. The table also shows the total receipts and outlays for the current fiscal year-to-date and the comparable prior fiscal year-to-date. This table includes total and subtotal rows that should be excluded when aggregating data. Some rows represent elements of the dataset's hierarchy, but are not assigned values. The classification_id for each of these elements can be used as the parent_id for underlying data elements to calculate their implied values. Subtotal rows are available to access this same information.                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                             |
| Monthly Treasury Statement (MTS)                                     | Trust Fund Impact on Budget Results and Investment Holdings                                          | `/v1/accounting/mts/mts_table_8`                                                             | This table shows the total receipts and outlays and the resulting surplus or deficit (shown on the table as excess) for the current month and the current fiscal year-to-date for all federal trust funds. The table also shows the totals for securities held as investments by the federal trust funds for the beginning of the fiscal year and the beginning and ending of the current accounting month. A trust fund is a type of account, designated by law, for receipts or offsetting receipts dedicated to specific purposes and the expenditure of these receipts. This table includes total and subtotal rows that should be excluded when aggregating data. Some rows represent elements of the dataset's hierarchy, but are not assigned values. The classification_id for each of these elements can be used as the parent_id for underlying data elements to calculate their implied values. Subtotal rows are available to access this same information.                                                                                                                                                                                                                                                                                                                                                                                         |
| Monthly Treasury Statement (MTS)                                     | Summary of Receipts by Source, and Outlays by Function of the U.S. Government                        | `/v1/accounting/mts/mts_table_9`                                                             | This summary table shows, for Budget Receipts, the total amount of activity for the current month, the current fiscal year-to-date, the comparable prior period year-to-date and the budgeted amount estimated for the current fiscal year for various types of receipts (i.e. individual income tax, corporate income tax, etc.). The Budget Outlays section of the table shows the total amount of activity for the current month, the current fiscal year-to-date, the comparable prior period year-to-date and the budgeted amount estimated for the current fiscal year for functions of the federal government. The table also shows the amounts for the budget/surplus deficit categorized as listed above. This table includes total and subtotal rows that should be excluded when aggregating data. Some rows represent elements of the dataset's hierarchy, but are not assigned values. The classification_id for each of these elements can be used as the parent_id for underlying data elements to calculate their implied values. Subtotal rows are available to access this same information.                                                                                                                                                                                                                                                  |
| Receipts by Department                                               | Receipts by Department                                                                               | `/v1/accounting/od/receipts_by_department`                                                   | The Receipts by Department table contains receipt line item amounts categorized by agency identifier, MAIN and SUB codes, and A code, if applicable                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                             |
| Record-Setting Treasury Securities Auction Data                      | Record-Setting Auction                                                                               | `/v2/accounting/od/record_setting_auction`                                                   | Rates, yields, offering, and auction dates of record high and low Treasury Bills, Notes, Bonds, Treasury Inflation-Protected Securities, Floating Rate Notes and Cash Management Bills.                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                         |
| Savings Bonds Securities Sold (Discontinued)                         | Savings Bonds Securities                                                                             | `/v1/accounting/od/slgs_savings_bonds`                                                       | Statistics on sold, redeemed, outstanding, and interest rates of non-marketable savings bonds.                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                  |
| Savings Bonds Value Files                                            | Savings Bonds Value Files                                                                            | `/v2/accounting/od/sb_value`                                                                 | The Savings Bond Value Files dataset is used by developers of bond pricing programs to update their systems with new redemption values for accrual savings bonds (Series E, EE, I & Savings Notes). The core data is the same as the Redemption Tables but there are differences in format, amount of data, and date range. The Savings Bonds Value Files dataset is meant for programmers and developers to read in redemption values without having to first convert PDFs.                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                    |
| Schedules of Federal Debt                                            | Schedules of Federal Debt by Month                                                                   | `/v1/accounting/od/schedules_fed_debt`                                                       | This table represents monthly activity for the Federal Debt Managed by the Bureau of the Fiscal Service separated by Held by the Public and Intragovernmental Debt Holdings totaled by Principal, Accrued Interest Payable, and Net Unamortized Premiums/Discounts. All figures are rounded to the nearest million.                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                             |
| Schedules of Federal Debt                                            | Schedules of Federal Debt, Fiscal Year-to-Date                                                       | `/v1/accounting/od/schedules_fed_debt_fytd`                                                  | This table represents fiscal year-to-date activity for the Federal Debt Managed by the Bureau of the Fiscal Service separated by Held by the Public and Intragovernmental Debt Holdings totaled by Principal, Accrued Interest Payable, and Net Unamortized Premiums/Discounts. All figures are rounded to the nearest million.                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                 |
| Schedules of Federal Debt by Day                                     | Daily Activity                                                                                       | `/v1/accounting/od/schedules_fed_debt_daily_activity`                                        | The Daily Activity Table provides daily changes in federal debt. The data notes whether the debt is debt held by the public or intragovernmental holdings. These two categories are further broken down into borrowing, repayment, interest, unamortized, and amortization amounts.                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                             |
| Schedules of Federal Debt by Day                                     | Daily Summary                                                                                        | `/v1/accounting/od/schedules_fed_debt_daily_summary`                                         | The Daily Activity Table provides daily federal debt balances. The data is broken out by debt held by the public and intragovernmental holdings amounts. These amounts include principal, accrued interest payable, and net unamortized. These amounts are in millions.                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                         |
| State and Local Government Series (SLGS) Daily Rate Table            | Demand Deposit Rate                                                                                  | `/v1/accounting/od/slgs_demand_deposit_rates`                                                | The Demand Deposit Rate table provides daily rates for demand deposit securities. The Demand Deposit rate covers all Demand Deposit securities issued, regardless of the redemption date.                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                       |
| State and Local Government Series (SLGS) Daily Rate Table            | Time Deposit Rate                                                                                    | `/v1/accounting/od/slgs_time_deposit_rates`                                                  | The Time Deposit Rate table provides daily rates for time deposit securities based on security life.                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                            |
| State and Local Government Series Securities (Non-Marketable)        | State and Local Government Series Securities (Non-Marketable)                                        | `/v1/accounting/od/slgs_securities`                                                          | Recap of State and Local Government Series (SLGS) transaction and balance activity updated daily including subscriptions, cancellations, issues, outstanding, and redemptions of non-marketable SLGS securities.                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                |
| TIPS and CPI Data                                                    | Reference CPI Numbers and Daily Index Ratios Summary Table                                           | `/v1/accounting/od/tips_cpi_data_summary`                                                    | The Reference CPI Numbers and Daily Index Ratios Summary Table contains overarching security information such as interest rate, security term, original auction date, and reference CPI on dated date among others.                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                             |
| TIPS and CPI Data                                                    | Reference CPI Numbers and Daily Index Ratios Details Table                                           | `/v1/accounting/od/tips_cpi_data_detail`                                                     | The Reference CPI Numbers and Daily Index Ratios Details Table contains the reference CPI and index ratio for each index date corresponding to each CUSIP.                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                      |
| Treasury Bulletin                                                    | PDO-1 - Offerings of Regular Weekly Treasury Bills                                                   | `/v1/accounting/tb/pdo1_offerings_regular_weekly_treasury_bills`                             | The Offerings of Regular Weekly Treasury Bills table presents the results of weekly auctions of 4, 8, 13, 17, and 26 week bills. Treasury bills mature each Thursday. Issues of 4 and 13 week bills are reopenings of 26 week bills. High rates on accepted tenders and the dollar value of total bids are presented, with the dollar value of awards made on both competitive and noncompetitive basis. To encourage the participation of individuals and smaller institutions, Treasury accepts noncompetitive tenders of up to $5 million in each auction of securities.                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                     |
| Treasury Bulletin                                                    | PDO-2 - Offerings of Marketable Securities Other than Regular Weekly Treasury Bills                  | `/v1/accounting/tb/pdo2_offerings_marketable_securities_other_regular_weekly_treasury_bills` | The Offerings of Marketable Securities Other than Regular Weekly Treasury Bills table lists the results of auctions of marketable securities, other than weekly bills, in chronological order over the past 2 years. Issues of cash management bills also are presented.                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                        |
| Treasury Bulletin                                                    | OFS-1 - Distribution of Federal Securities by Class of Investors and Type of Issues                  | `/v1/accounting/tb/ofs1_distribution_federal_securities_class_investors_type_issues`         | The Distribution of Federal Securities by Class of Investors and Type of Issues table presents Treasury marketable andnonmarketable securities and debt issued by other Federal agencies held by Government accounts, the FRBs, and private investors. Social Security and Federal retirement trust fund investments comprise much of the Government account holdings. The FRBs acquire Treasury securities in the market as a means of executing monetary policy.                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                              |
| Treasury Bulletin                                                    | OFS-2 - Estimated Ownership of U.S. Treasury Securities                                              | `/v1/accounting/tb/ofs2_estimated_ownership_treasury_securities`                             | The Estimated Ownership of U.S. Treasury Securities table presents the estimated ownership of U.S. Treasury securities. Information is primarily obtained from the Federal Reserve Board of Governors Flow of Funds data, Table L209. State, local, andforeign holdings include special issues of nonmarketable securities to municipal entities and foreign official accounts. They also include municipal, foreign official, and private holdings of marketable Treasury securities.                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                          |
| Treasury Bulletin                                                    | USCC-1 - Amounts Outstanding and in Circulation                                                      | `/v1/accounting/tb/uscc1_amounts_outstanding_circulation`                                    | The USCC-1 - Amounts Outstanding and in Circulation table informs the public of the total face value of currency and coin used as a medium of exchange that is in circulation. It defines the total amount of currency and coin outstanding and the portion deemed to be in circulation. It includes some old and current rare issues that do not circulate or that may do so to a limited extent. Treasury includes them in the statement because the issues were originally intended for general circulation. The USCC comes from monthly reports compiled by Treasury offices, U.S. Mint offices, the Federal Reserve banks (FRBs), and the Federal Reserve Board.                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                           |
| Treasury Bulletin                                                    | USCC-2 - Amounts Outstanding and in Circulation                                                      | `/v1/accounting/tb/uscc2_amounts_outstanding_circulation`                                    | The USCC-2 - Amounts Outstanding and in Circulation table provides a description of the various issues of paper money by denomination. Partial notes represents value of certain partial denominations not presented for redemption. It also gives an estimated average of currency and coin held by each individual, using estimates of population from the Bureau of the Census. The USCC comes from monthly reports compiled by Treasury offices, U.S. Mint offices, the Federal Reserve banks (FRBs), and the Federal Reserve Board.                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                        |
| Treasury Bulletin                                                    | FCP-1 - Weekly Report of Major Market Participants                                                   | `/v1/accounting/tb/fcp1_weekly_report_major_market_participants`                             | The Foreign Currency Positions Weekly Report of Major Market Participants contains foreign currency holdings of large foreign exchange market participants. This table presents the currency data reported weekly by major market participants. This table provides information on positions in derivative instruments, such as foreign exchange futures and options that are increasingly used in establishing foreign exchange positions. Weekly reports must be filed throughout the calendar year by major foreign exchange market participants, which are defined as market participants with more than $50 billion equivalent in foreign exchange contracts on the last business day of any calendar quarter during the previous year (end March, September, September, or December). Such contracts include the amounts of foreign exchange spotcontracts bought and sold, foreign exchange forward contracts bought and sold, foreign exchange futures bought and sold, and one half the notional amount of foreign exchange options bought and sold. The information in the table is based on the reports referenced in this Introduction: Foreign Currency Positions and is not audited by the Federal Reserve banks or the Treasury Department. Please note that these amounts are reported in the foreign currency specified.                       |
| Treasury Bulletin                                                    | FCP-2 - Monthly Report of Major Market Participants                                                  | `/v1/accounting/tb/fcp2_monthly_report_major_market_participants`                            | The Foreign Currency Positions Monthly Report of Major Market Participants contains foreign currency holdings of large foreign exchange market participants. This table presents more detailed currency data of major market participants, based on monthly reports. This table provides information on positions in derivative instruments, such as foreign exchange futures and options that are increasingly used in establishing foreign exchange positions. Monthly reports must be filed throughout the calendar year by major foreign exchange market participants, which are defined as market participants with more than $50 billion equivalent in foreign exchange contracts on the last business day of any calendar quarter during the previous year (end March, September, September, or December). Such contracts include the amounts of foreign exchange spotcontracts bought and sold, foreign exchange forward contracts bought and sold, foreign exchange futures bought and sold, and one half the notional amount of foreign exchange options bought and sold. The information in the table is based on the reports referenced in this Introduction: Foreign Currency Positions and is not audited by the Federal Reserve banks or the Treasury Department. Please note that these amounts are reported in the foreign currency specified. |
| Treasury Bulletin                                                    | FCP-3 - Quarterly Report of Large Market Participants                                                | `/v1/accounting/tb/fcp3_quarterly_report_large_market_participants`                          | The Foreign Currency Positions Quarterly Report of Large Market Participants contains foreign currency holdings of large foreign exchange market participants. This table presents quarterly consolidated currency data reported by large market participants that do not file weekly reports. This table provides information on positions in derivative instruments, such as foreign exchange futures and options that are increasingly used in establishing foreign exchange positions. A quarterly report must be filed throughout the calendar year by each foreign exchange market participant that had more than $5 billion equivalent in foreign exchange contracts on the last business day of any quarter the previous year (end March, June, September, or December). The information in the table is based on the reports referenced in this Introduction: Foreign Currency Positions and is not audited by the Federal Reserve banks or the Treasury Department. Please note that these amounts are reported in the foreign currency specified.                                                                                                                                                                                                                                                                                                    |
| Treasury Bulletin                                                    | ESF-1 - Balances                                                                                     | `/v1/accounting/tb/esf1_balances`                                                            | The Exchange Stabilization Fund Balances (ESF-1) table presents the assets, liabilities, and net position of the fund. The figures are in U.S. dollars. Amounts and transactions pertaining to foreign currencies and special drawing rights, an international reserve asset created by the International Monetary Fund (IMF) to supplement its member countries position in the Fund, have been converted to U.S. dollars based on current exchange rates computed according to the accrual method of accounting. Investments and loans receivable are reported at fair value. Unexpended Appropriations Funds from Dedicated Collections represents the original capital appropriated to the fund by Congress of $2 billion, minus a subsequent transfer of $1.8 billion to pay for the initial U.S. quota subscription to the International Monetary Fund. Unexpended Appropriations Funds from Other than Dedicated Collections represents the amount appropriated under the CARES Act Sec 4027 minus transfers, expenditures for administrative and subsidy costs, and rescissions resulting from the passage of the Consolidated Appropriations Act, 2021 and the Infrastructure Investment and Jobs Act, 2021. Conversion gains and losses are reflected in the cumulative net income -+ or loss -- account.                                             |
| Treasury Bulletin                                                    | ESF-2 - Statement of Net Cost                                                                        | `/v1/accounting/tb/esf2_statement_net_cost`                                                  | The Exchange Stabilization Fund Statement of Net Cost (ESF-2) table shows net cost from operations for the current quarter and year-to-date. Figures are in U.S. dollars computed according to the accrual method. Gains-+ or loss -- on foreign exchange includes both realized and unrealized gains or losses. Adjustment for change in valuation of SDR holdings and allocations reflects net gain or loss on revaluation of SDR holdings and allocations for the quarter. CARES Act related administrative costs incurred in connection with the loans, and other investments are accrued.                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                  |
| Treasury Bulletin                                                    | FFO-5 - Internal Revenue Receipts by State                                                           | `/v1/accounting/tb/ffo5_internal_revenue_by_state`                                           | The Internal Revenue Receipts by State table summarizes internal revenue receipts by states and by type of tax. Amounts reported are collections made in a fiscal year. They span several tax liability years because they consist of prepayments (estimated tax payments and taxes withheld by employers for individual income and Social Security taxes), payments made with tax returns and subsequent payments made after tax returns are due or are filed (that is, payments with delinquent returns or on delinquent accounts). Amounts are reported based on the primary filing address provided by each taxpayer or reporting entity. For multistate corporations, the address may reflect only the district where such a corporation reported its taxes from a principal office rather than other districts where income was earned or where individual income and Social Security taxes were withheld. In addition, an individual may reside in one district and work in another.                                                                                                                                                                                                                                                                                                                                                                     |
| Treasury Bulletin                                                    | FFO-6 - Customs and Border Protection Collection of Duties, Taxes, and Fees by Districts and Ports   | `/v1/accounting/tb/ffo6_customs_border_protection_collections`                               | The Customs and Border Protection Collection of Duties, Taxes, and Fees by Districts and Ports table includes customs collection of duties, taxes, and fees by districts and ports. Reported by U.S. Customs and Border Protection.                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                             |
| Treasury Bulletin: Trust Fund Reports                                | Airport and Airway Trust Fund Results of Operations                                                  | `/v1/accounting/od/airport_airway_trust_fund_results`                                        | The Airport and Airway Trust Fund Results of Operations table contains receipt, refund, and expense amounts along with a beginning and end of fiscal year balance, as reported in the Treasury Bulletin. The Airport and Airway Trust Fund was established on the books of the Department of Treasury in fiscal year 1971, according to provisions of the Airport and Airway Revenue Act of 1970 [49 United States Code 1742(a), repealed]. The Tax Equity and Fiscal Responsibility Act of 1982 (Public Law 97-248, dated September 3, 1982) reestablished the trust fund in the Internal Revenue Code (26 United States Code 9502) effective September 1, 1982. Treasury transfers from the general fund to the trust fund amounts equivalent to the taxes received from transportation of persons and property by air, gasoline and jet fuel used in commercial and noncommercial aircraft, and an international arrival and departure tax. Please see the published Treasury Bulletin for more information regarding this fund.                                                                                                                                                                                                                                                                                                                             |
| Treasury Bulletin: Trust Fund Reports                                | Airport and Airway Trust Fund Expected Condition and Results of Operations                           | `/v1/accounting/od/airport_airway_trust_fund_expected`                                       | The Airport and Airway Trust Fund Expected Condition and Results of Operations table contains receipt and expense amounts along with a beginning and end of fiscal year balance for the next five years, as reported in the Treasury Bulletin. The Airport and Airway Trust Fund was established on the books of the Department of Treasury in fiscal year 1971, according to provisions of the Airport and Airway Revenue Act of 1970 [49 United States Code 1742(a), repealed]. The Tax Equity and Fiscal Responsibility Act of 1982 (Public Law 97-248, dated September 3, 1982) reestablished the trust fund in the Internal Revenue Code (26 United States Code 9502) effective September 1, 1982. Treasury transfers from the general fund to the trust fund amounts equivalent to the taxes received from transportation of persons and property by air, gasoline and jet fuel used in commercial and noncommercial aircraft, and an international arrival and departure tax. Please see the published Treasury Bulletin for more information regarding this fund.                                                                                                                                                                                                                                                                                       |
| Treasury Bulletin: Trust Fund Reports                                | Uranium Enrichment Decontamination and Decommissioning Fund Results of Operations                    | `/v1/accounting/od/uranium_enrichment_decontamination_decommissioning_fund_results`          | The Uranium Enrichment Decontamination and Decommissioning Fund Results of Operations table contains recipt, transfer, and outlay amounts along with a beginning and end of fiscal year balance, as reported in the Treasury Bulletin. The Uranium Enrichment Decontamination and Decommissioning Fund was established on the books of the Treasury in fiscal year (FY) 1993, in accordance with provisions of the Energy Policy Act of 1992 (42 United States Code 2297g). Please see the published Treasury Bulletin for more information regarding this fund.                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                |
| Treasury Bulletin: Trust Fund Reports                                | Uranium Enrichment Decontamination and Decommissioning Fund Expected Cond. and Results of Operations | `/v1/accounting/od/uranium_enrichment_decontamination_decommissioning_fund_expected`         | The Uranium Enrichment Decontamination and Decommissioning Fund Expected Condition and Results of Operations table contains recipt and outlay amounts along with a beginning and end of fiscal year balance for the next five years, as reported in the Treasury Bulletin. The Uranium Enrichment Decontamination and Decommissioning Fund was established on the books of the Treasury in fiscal year (FY) 1993, in accordance with provisions of the Energy Policy Act of 1992 (42 United States Code 2297g). Please see the published Treasury Bulletin for more information regarding this fund.                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                            |
| Treasury Bulletin: Trust Fund Reports                                | Black Lung Disability Trust Fund Results of Operations                                               | `/v1/accounting/od/black_lung_disability_trust_fund_results`                                 | The Black Lung Disability Trust Fund Results of Operations table contains receipt, outlay, and expense amounts along with a beginning and end of fiscal year balance, as reported in the Treasury Bulletin. The Black Lung Disability Trust Fund was established on the books of the Treasury in fiscal year 1978 according to the Black Lung Benefits Revenue Act of 1977 (Public Law 95-227). The Black Lung Benefits Revenue Act of 1981 (Public Law 97-119) reestablished the fund in the Internal Revenue Code (IRC), 26 United States Code 9501. Please see the published Treasury Bulletin for more information regarding this fund.                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                     |
| Treasury Bulletin: Trust Fund Reports                                | Black Lung Disability Trust Fund Expected Condition and Results of Operations                        | `/v1/accounting/od/black_lung_disability_trust_fund_expected`                                | The Black Lung Disability Trust Fund Expected Condition and Results of Operations table contains receipt, outlay, and expense amounts along with a beginning and end of fiscal year balance for the next five years, as reported in the Treasury Bulletin. The Black Lung Disability Trust Fund was established on the books of the Treasury in fiscal year 1978 according to the Black Lung Benefits Revenue Act of 1977 (Public Law 95-227). The Black Lung Benefits Revenue Act of 1981 (Public Law 97-119) reestablished the fund in the Internal Revenue Code (IRC), 26 United States Code 9501. Please see the published Treasury Bulletin for more information regarding this fund.                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                      |
| Treasury Bulletin: Trust Fund Reports                                | Harbor Maintenance Trust Fund Results of Operation                                                   | `/v1/accounting/od/harbor_maintenance_trust_fund_results`                                    | The Harbor Maintenance Trust Fund Results of Operation table contains receipt and expense amounts along with a beginning and end of fiscal year balance, as reported in the Treasury Bulletin. The Harbor Maintenance Trust Fund was established on the books of the Treasury on April 1, 1987, according to the Water Resources Development Act of 1986 (Public Law 99-662, November 17, 1986) (26 United States Code 9505). Please see the published Treasury Bulletin for more information regarding this fund.                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                              |
| Treasury Bulletin: Trust Fund Reports                                | Harbor Maintenance Trust Fund Expected Condition and Results of Operations                           | `/v1/accounting/od/harbor_maintenance_trust_fund_expected`                                   | The Harbor Maintenance Trust Fund Expected Condition and Results of Operations table contains receipt and outlay amounts along with a beginning and end of fiscal year balance for the next five years, as reported in the Treasury Bulletin. The Harbor Maintenance Trust Fund was established on the books of the Treasury on April 1, 1987, according to the Water Resources Development Act of 1986 (Public Law 99-662, November 17, 1986) (26 United States Code 9505). Please see the published Treasury Bulletin for more information regarding this fund.                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                               |
| Treasury Bulletin: Trust Fund Reports                                | Hazardous Substance Superfund Results of Operations                                                  | `/v1/accounting/od/hazardous_substance_superfund_results`                                    | The Hazardous Substance Superfund Results of Operations table contains receipts and expense amounts along with a beginning and end of fiscal year balance, as reported in the Treasury Bulletin. The Hazardous Substance Response Trust Fund was established on the books of the Treasury in fiscal year 1981, in accordance with section 221 of the Hazardous Substance Response Revenue Act of 1980 [42 United States Code 9631(a), repealed]. The trust fund was renamed the Hazardous Substance Superfund (Superfund) and relocated in accordance with section 517 of the Superfund Amendments and Reauthorization Act of 1986 [Public Law 99-499, dated October 17, 1986 (26 United States Code 9507)]. Please see the published Treasury Bulletin for more information regarding this fund.                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                               |
| Treasury Bulletin: Trust Fund Reports                                | Hazardous Substance Superfund Expected Condition and Results of Operations                           | `/v1/accounting/od/hazardous_substance_superfund_expected`                                   | The Hazardous Substance Superfund Expected Condition and Results of Operations table contains receipts, appropriations, and expense amounts along with a beginning and end of fiscal year balance for the next five years, as reported in the Treasury Bulletin. The Hazardous Substance Response Trust Fund was established on the books of the Treasury in fiscal year 1981, in accordance with section 221 of the Hazardous Substance Response Revenue Act of 1980 [42 United States Code 9631(a), repealed]. The trust fund was renamed the Hazardous Substance Superfund (Superfund) and relocated in accordance with section 517 of the Superfund Amendments and Reauthorization Act of 1986 [Public Law 99-499, dated October 17, 1986 (26 United States Code 9507)]. Please see the published Treasury Bulletin for more information regarding this fund.                                                                                                                                                                                                                                                                                                                                                                                                                                                                                               |
| Treasury Bulletin: Trust Fund Reports                                | Highway Trust Fund Results of Operations                                                             | `/v1/accounting/od/highway_trust_fund_results`                                               | The Highway Trust Fund Results of Operations table contains receipt and expense amounts along with a beginning and end of fiscal year balance, as reported in the Treasury Bulletin. The Highway Trust Fund was established on the books of the Treasury in fiscal year 1957, according to provisions of the Highway Revenue Act of 1956 (Act of June 29, 1956, chapter 462, section 209). It has been amended and extended by various highway surface transportation and other acts since 1959. Please see the published Treasury Bulletin for more information regarding this fund.                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                           |
| Treasury Bulletin: Trust Fund Reports                                | Highway Trust Fund Expected Condition and Results of Operations                                      | `/v1/accounting/od/highway_trust_fund_expected`                                              | The Highway Trust Fund Expected Condition and Results of Operations table contains receipt and expense amounts along with a beginning and end of fiscal year balance for the next five years, as reported in the Treasury Bulletin. The Highway Trust Fund was established on the books of the Treasury in fiscal year 1957, according to provisions of the Highway Revenue Act of 1956 (Act of June 29, 1956, chapter 462, section 209). It has been amended and extended by various highway surface transportation and other acts since 1959. Please see the published Treasury Bulletin for more information regarding this fund.                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                            |
| Treasury Bulletin: Trust Fund Reports                                | Highway Trust Fund                                                                                   | `/v1/accounting/od/highway_trust_fund`                                                       | The Highway Trust Fund table contains commitment, cash balance, unfunded authorizations, and 48 month revenue estimate amounts, as reported in the Treasury Bulletin. The Highway Trust Fund was established on the books of the Treasury in fiscal year 1957, according to provisions of the Highway Revenue Act of 1956 (Act of June 29, 1956, chapter 462, section 209). It has been amended and extended by various highway surface transportation and other acts since 1959. Please see the published Treasury Bulletin for more information regarding this fund.                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                          |

## API Documentation
- [Fiscal Data (Treasury.gov) API Documentation](https://fiscaldata.treasury.gov/api-documentation/)